# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


def fmt(docstr):
    """Format a docstring for use as documentation in sample config."""
    # Replace newlines with spaces, as docstrings contain literal newlines that
    # should not be rendered into the sample configuration file (instead, line
    # wrappings should be applied automatically).
    docstr = docstr.replace('\n', ' ')

    # Because it's common for docstrings to begin and end with a newline, there
    # is now whitespace at the beginning and end of the documentation as a side
    # effect of replacing newlines with spaces.
    docstr = docstr.strip()

    return docstr



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import http.client
import uuid

from oslo_serialization import jsonutils

from keystone.common.policies import grant as gp
from keystone.common import provider_api
import keystone.conf
from keystone.tests.common import auth as common_auth
from keystone.tests import unit
from keystone.tests.unit import base_classes
from keystone.tests.unit import ksfixtures
from keystone.tests.unit.ksfixtures import temporaryfile

CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


class _SystemUserGrantTests:
    def test_can_list_grants_for_user_on_project(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=user['id'],
            project_id=project['id'],
        )

        with self.test_client() as c:
            r = c.get(
                '/v3/projects/{}/users/{}/roles'.format(
                    project['id'], user['id']
                ),
                headers=self.headers,
            )
            self.assertEqual(1, len(r.json['roles']))

    def test_can_list_grants_for_user_on_domain(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        )

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=user['id'],
            domain_id=domain['id'],
        )

        with self.test_client() as c:
            r = c.get(
                '/v3/domains/{}/users/{}/roles'.format(
                    domain['id'], user['id']
                ),
                headers=self.headers,
            )
            self.assertEqual(1, len(r.json['roles']))

    def test_can_list_grants_for_group_on_project(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=CONF.identity.default_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            group_id=group['id'],
            project_id=project['id'],
        )

        with self.test_client() as c:
            r = c.get(
                '/v3/projects/{}/groups/{}/roles'.format(
                    project['id'], group['id']
                ),
                headers=self.headers,
            )
            self.assertEqual(1, len(r.json['roles']))

    def test_can_list_grants_for_group_on_domain(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=CONF.identity.default_domain_id)
        )

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            group_id=group['id'],
            domain_id=domain['id'],
        )

        with self.test_client() as c:
            r = c.get(
                '/v3/domains/{}/groups/{}/roles'.format(
                    domain['id'], group['id']
                ),
                headers=self.headers,
            )
            self.assertEqual(1, len(r.json['roles']))

    def test_can_check_grant_for_user_on_project(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=user['id'],
            project_id=project['id'],
        )

        with self.test_client() as c:
            c.get(
                '/v3/projects/{}/users/{}/roles/{}'.format(
                    project['id'], user['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.NO_CONTENT,
            )

    def test_can_check_grant_for_user_on_domain(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        )

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=user['id'],
            domain_id=domain['id'],
        )

        with self.test_client() as c:
            c.get(
                '/v3/domains/{}/users/{}/roles/{}'.format(
                    domain['id'], user['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.NO_CONTENT,
            )

    def test_can_check_grant_for_group_on_project(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=CONF.identity.default_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            group_id=group['id'],
            project_id=project['id'],
        )

        with self.test_client() as c:
            c.get(
                '/v3/projects/{}/groups/{}/roles/{}'.format(
                    project['id'],
                    group['id'],
                    self.bootstrapper.reader_role_id,
                ),
                headers=self.headers,
                expected_status_code=http.client.NO_CONTENT,
            )

    def test_can_check_grant_for_group_on_domain(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=CONF.identity.default_domain_id)
        )

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            group_id=group['id'],
            domain_id=domain['id'],
        )

        with self.test_client() as c:
            c.get(
                '/v3/domains/{}/groups/{}/roles/{}'.format(
                    domain['id'], group['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.NO_CONTENT,
            )


class _SystemMemberAndReaderGrantTests:
    def test_cannot_create_grant_for_user_on_project(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        with self.test_client() as c:
            c.put(
                '/v3/projects/{}/users/{}/roles/{}'.format(
                    project['id'], user['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_create_grant_for_user_on_domain(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        )

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        with self.test_client() as c:
            c.put(
                '/v3/domains/{}/users/{}/roles/{}'.format(
                    domain['id'], user['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_create_grant_for_group_on_project(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=CONF.identity.default_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        with self.test_client() as c:
            c.put(
                '/v3/projects/{}/groups/{}/roles/{}'.format(
                    project['id'],
                    group['id'],
                    self.bootstrapper.reader_role_id,
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_create_grant_for_group_on_domain(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=CONF.identity.default_domain_id)
        )

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        with self.test_client() as c:
            c.put(
                '/v3/domains/{}/groups/{}/roles/{}'.format(
                    domain['id'], group['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_revoke_grant_from_user_on_project(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=user['id'],
            project_id=project['id'],
        )

        with self.test_client() as c:
            c.delete(
                '/v3/projects/{}/users/{}/roles/{}'.format(
                    project['id'], user['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_revoke_grant_from_user_on_domain(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        )

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=user['id'],
            domain_id=domain['id'],
        )

        with self.test_client() as c:
            c.delete(
                '/v3/domains/{}/users/{}/roles/{}'.format(
                    domain['id'], user['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_revoke_grant_from_group_on_project(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=CONF.identity.default_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            group_id=group['id'],
            project_id=project['id'],
        )

        with self.test_client() as c:
            c.delete(
                '/v3/projects/{}/groups/{}/roles/{}'.format(
                    project['id'],
                    group['id'],
                    self.bootstrapper.reader_role_id,
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_revoke_grant_from_group_on_domain(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=CONF.identity.default_domain_id)
        )

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            group_id=group['id'],
            domain_id=domain['id'],
        )

        with self.test_client() as c:
            c.delete(
                '/v3/domains/{}/groups/{}/roles/{}'.format(
                    domain['id'], group['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )


class _DomainUserTests:
    def test_can_list_grants_for_user_on_project(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=self.domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=user['id'],
            project_id=project['id'],
        )

        with self.test_client() as c:
            r = c.get(
                '/v3/projects/{}/users/{}/roles'.format(
                    project['id'], user['id']
                ),
                headers=self.headers,
            )
            self.assertEqual(1, len(r.json['roles']))

    def test_can_list_grants_for_user_on_domain(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=user['id'],
            domain_id=self.domain_id,
        )

        with self.test_client() as c:
            r = c.get(
                '/v3/domains/{}/users/{}/roles'.format(
                    self.domain_id, user['id']
                ),
                headers=self.headers,
            )
            self.assertEqual(1, len(r.json['roles']))

    def test_can_list_grants_for_group_on_project(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=self.domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=self.domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            group_id=group['id'],
            project_id=project['id'],
        )

        with self.test_client() as c:
            r = c.get(
                '/v3/projects/{}/groups/{}/roles'.format(
                    project['id'], group['id']
                ),
                headers=self.headers,
            )
            self.assertEqual(1, len(r.json['roles']))

    def test_can_list_grants_for_group_on_domain(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=self.domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            group_id=group['id'],
            domain_id=self.domain_id,
        )

        with self.test_client() as c:
            r = c.get(
                '/v3/domains/{}/groups/{}/roles'.format(
                    self.domain_id, group['id']
                ),
                headers=self.headers,
            )
            self.assertEqual(1, len(r.json['roles']))

    def test_can_check_grant_for_user_on_project(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=self.domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=user['id'],
            project_id=project['id'],
        )

        with self.test_client() as c:
            c.get(
                '/v3/projects/{}/users/{}/roles/{}'.format(
                    project['id'], user['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.NO_CONTENT,
            )

    def test_can_check_grant_for_user_on_domain(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=user['id'],
            domain_id=self.domain_id,
        )

        with self.test_client() as c:
            c.get(
                '/v3/domains/{}/users/{}/roles/{}'.format(
                    self.domain_id,
                    user['id'],
                    self.bootstrapper.reader_role_id,
                ),
                headers=self.headers,
                expected_status_code=http.client.NO_CONTENT,
            )

    def test_can_check_grant_for_group_on_project(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=self.domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=self.domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            group_id=group['id'],
            project_id=project['id'],
        )

        with self.test_client() as c:
            c.get(
                '/v3/projects/{}/groups/{}/roles/{}'.format(
                    project['id'],
                    group['id'],
                    self.bootstrapper.reader_role_id,
                ),
                headers=self.headers,
                expected_status_code=http.client.NO_CONTENT,
            )

    def test_can_check_grant_for_group_on_domain(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=self.domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            group_id=group['id'],
            domain_id=self.domain_id,
        )

        with self.test_client() as c:
            c.get(
                '/v3/domains/{}/groups/{}/roles/{}'.format(
                    self.domain_id,
                    group['id'],
                    self.bootstrapper.reader_role_id,
                ),
                headers=self.headers,
                expected_status_code=http.client.NO_CONTENT,
            )

    def test_cannot_list_grants_for_user_other_domain_on_project_own_domain(
        self,
    ):  # noqa: E501
        user_domain_id = CONF.identity.default_domain_id
        project_domain_id = self.domain_id

        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=user_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=project_domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=user['id'],
            project_id=project['id'],
        )

        with self.test_client() as c:
            c.get(
                '/v3/projects/{}/users/{}/roles'.format(
                    project['id'], user['id']
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_list_grants_for_user_own_domain_on_project_other_domain(
        self,
    ):  # noqa: E501
        user_domain_id = self.domain_id
        project_domain_id = CONF.identity.default_domain_id

        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=user_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=project_domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=user['id'],
            project_id=project['id'],
        )

        with self.test_client() as c:
            c.get(
                '/v3/projects/{}/users/{}/roles'.format(
                    project['id'], user['id']
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_list_grants_for_user_own_domain_on_other_domain(self):
        user_domain_id = self.domain_id
        domain_id = CONF.identity.default_domain_id

        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=user_domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=user['id'],
            domain_id=domain_id,
        )

        with self.test_client() as c:
            c.get(
                '/v3/domains/{}/users/{}/roles'.format(domain_id, user['id']),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_list_grants_for_user_other_domain_on_own_domain(self):
        user_domain_id = CONF.identity.default_domain_id
        domain_id = self.domain_id

        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=user_domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=user['id'],
            domain_id=domain_id,
        )

        with self.test_client() as c:
            c.get(
                '/v3/domains/{}/users/{}/roles'.format(domain_id, user['id']),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_list_grants_for_group_other_domain_on_project_own_domain(
        self,
    ):  # noqa: E501
        group_domain_id = CONF.identity.default_domain_id
        project_domain_id = self.domain_id

        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=group_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=project_domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            group_id=group['id'],
            project_id=project['id'],
        )

        with self.test_client() as c:
            c.get(
                '/v3/projects/{}/groups/{}/roles'.format(
                    project['id'], group['id']
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_list_grants_for_group_own_domain_on_project_other_domain(
        self,
    ):  # noqa: E501
        group_domain_id = self.domain_id
        project_domain_id = CONF.identity.default_domain_id

        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=group_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=project_domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            group_id=group['id'],
            project_id=project['id'],
        )

        with self.test_client() as c:
            c.get(
                '/v3/projects/{}/groups/{}/roles'.format(
                    project['id'], group['id']
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_list_grants_for_group_own_domain_on_other_domain(self):
        group_domain_id = self.domain_id
        domain_id = CONF.identity.default_domain_id

        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=group_domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            group_id=group['id'],
            domain_id=domain_id,
        )

        with self.test_client() as c:
            c.get(
                '/v3/domains/{}/groups/{}/roles'.format(
                    domain_id, group['id']
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_list_grants_for_group_other_domain_on_own_domain(self):
        group_domain_id = CONF.identity.default_domain_id
        domain_id = self.domain_id

        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=group_domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            group_id=group['id'],
            domain_id=domain_id,
        )

        with self.test_client() as c:
            c.get(
                '/v3/domains/{}/groups/{}/roles'.format(
                    domain_id, group['id']
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_check_grant_for_user_other_domain_on_project_own_domain(
        self,
    ):  # noqa: E501
        user_domain_id = CONF.identity.default_domain_id
        project_domain_id = self.domain_id

        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=user_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=project_domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=user['id'],
            project_id=project['id'],
        )

        with self.test_client() as c:
            c.get(
                '/v3/projects/{}/users/{}/roles/{}'.format(
                    project['id'], user['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_check_grant_for_user_own_domain_on_project_other_domain(
        self,
    ):  # noqa: E501
        user_domain_id = self.domain_id
        project_domain_id = CONF.identity.default_domain_id

        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=user_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=project_domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=user['id'],
            project_id=project['id'],
        )

        with self.test_client() as c:
            c.get(
                '/v3/projects/{}/users/{}/roles/{}'.format(
                    project['id'], user['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_check_grant_for_user_own_domain_on_project_own_domain_with_role_other_domain(
        self,
    ):  # noqa: E501
        user_domain_id = self.domain_id
        project_domain_id = self.domain_id
        role_domain_id = CONF.identity.default_domain_id

        role = PROVIDERS.role_api.create_role(
            uuid.uuid4().hex, unit.new_role_ref(domain_id=role_domain_id)
        )

        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=user_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=project_domain_id)
        )

        # NOTE(cmurphy) the grant for a domain-specific role cannot be created
        # for a project in a different domain, so we don't try to create it,
        # but we still need to test that checking the role results in a 403 and
        # not a 404

        with self.test_client() as c:
            c.get(
                '/v3/projects/{}/users/{}/roles/{}'.format(
                    project['id'], user['id'], role['id']
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_check_grant_for_user_own_domain_on_other_domain(self):
        user_domain_id = self.domain_id
        domain_id = CONF.identity.default_domain_id

        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=user_domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=user['id'],
            domain_id=domain_id,
        )

        with self.test_client() as c:
            c.get(
                '/v3/domains/{}/users/{}/roles/{}'.format(
                    domain_id, user['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_check_grant_for_user_other_domain_on_own_domain(self):
        user_domain_id = CONF.identity.default_domain_id
        domain_id = self.domain_id

        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=user_domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=user['id'],
            domain_id=domain_id,
        )

        with self.test_client() as c:
            c.get(
                '/v3/domains/{}/users/{}/roles/{}'.format(
                    domain_id, user['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_check_grant_for_user_own_domain_on_own_domain_with_role_other_domain(
        self,
    ):  # noqa: E501
        user_domain_id = self.domain_id
        domain_id = self.domain_id
        role_domain_id = CONF.identity.default_domain_id

        role = PROVIDERS.role_api.create_role(
            uuid.uuid4().hex, unit.new_role_ref(domain_id=role_domain_id)
        )

        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=user_domain_id)
        )

        # NOTE(cmurphy) the grant for a domain-specific role cannot be created
        # for a project in a different domain, so we don't try to create it,
        # but we still need to test that checking the role results in a 403 and
        # not a 404

        with self.test_client() as c:
            c.get(
                '/v3/domains/{}/users/{}/roles/{}'.format(
                    domain_id, user['id'], role['id']
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_check_grant_for_group_other_domain_on_project_own_domain(
        self,
    ):  # noqa: E501
        group_domain_id = CONF.identity.default_domain_id
        project_domain_id = self.domain_id

        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=group_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=project_domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            group_id=group['id'],
            project_id=project['id'],
        )

        with self.test_client() as c:
            c.get(
                '/v3/projects/{}/groups/{}/roles/{}'.format(
                    project['id'],
                    group['id'],
                    self.bootstrapper.reader_role_id,
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_check_grant_for_group_own_domain_on_project_other_domain(
        self,
    ):  # noqa: E501
        group_domain_id = self.domain_id
        project_domain_id = CONF.identity.default_domain_id

        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=group_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=project_domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            group_id=group['id'],
            project_id=project['id'],
        )

        with self.test_client() as c:
            c.get(
                '/v3/projects/{}/groups/{}/roles/{}'.format(
                    project['id'],
                    group['id'],
                    self.bootstrapper.reader_role_id,
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_check_grant_for_group_own_domain_on_project_own_domain_with_role_other_domain(
        self,
    ):  # noqa: E501
        group_domain_id = self.domain_id
        project_domain_id = CONF.identity.default_domain_id
        role_domain_id = CONF.identity.default_domain_id

        role = PROVIDERS.role_api.create_role(
            uuid.uuid4().hex, unit.new_role_ref(domain_id=role_domain_id)
        )

        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=group_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=project_domain_id)
        )

        # NOTE(cmurphy) the grant for a domain-specific role cannot be created
        # for a project in a different domain, so we don't try to create it,
        # but we still need to test that checking the role results in a 403 and
        # not a 404

        with self.test_client() as c:
            c.get(
                '/v3/projects/{}/groups/{}/roles/{}'.format(
                    project['id'], group['id'], role['id']
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_check_grant_for_group_own_domain_on_other_domain(self):
        group_domain_id = self.domain_id
        domain_id = CONF.identity.default_domain_id

        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=group_domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            group_id=group['id'],
            domain_id=domain_id,
        )

        with self.test_client() as c:
            c.get(
                '/v3/domains/{}/groups/{}/roles/{}'.format(
                    domain_id, group['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_check_grant_for_group_other_domain_on_own_domain(self):
        group_domain_id = CONF.identity.default_domain_id
        domain_id = self.domain_id

        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=group_domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            group_id=group['id'],
            domain_id=domain_id,
        )

        with self.test_client() as c:
            c.get(
                '/v3/domains/{}/groups/{}/roles/{}'.format(
                    domain_id, group['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_check_grant_for_group_own_domain_on_own_domain_with_role_other_domain(
        self,
    ):  # noqa: E501
        group_domain_id = self.domain_id
        domain_id = self.domain_id
        role_domain_id = CONF.identity.default_domain_id

        role = PROVIDERS.role_api.create_role(
            uuid.uuid4().hex, unit.new_role_ref(domain_id=role_domain_id)
        )

        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=group_domain_id)
        )

        # NOTE(cmurphy) the grant for a domain-specific role cannot be created
        # for a project in a different domain, so we don't try to create it,
        # but we still need to test that checking the role results in a 403 and
        # not a 404

        with self.test_client() as c:
            c.get(
                '/v3/domains/{}/groups/{}/roles/{}'.format(
                    domain_id, group['id'], role['id']
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_create_grant_for_user_other_domain_on_project_own_domain(
        self,
    ):  # noqa: E501
        user_domain_id = CONF.identity.default_domain_id
        project_domain_id = self.domain_id

        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=user_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=project_domain_id)
        )

        with self.test_client() as c:
            c.put(
                '/v3/projects/{}/users/{}/roles/{}'.format(
                    project['id'], user['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_create_grant_for_user_own_domain_on_project_other_domain(
        self,
    ):  # noqa: E501
        user_domain_id = self.domain_id
        project_domain_id = CONF.identity.default_domain_id

        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=user_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=project_domain_id)
        )

        with self.test_client() as c:
            c.put(
                '/v3/projects/{}/users/{}/roles/{}'.format(
                    project['id'], user['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_create_grant_for_user_own_domain_on_project_own_domain_with_role_other_domain(
        self,
    ):  # noqa: E501
        user_domain_id = self.domain_id
        project_domain_id = self.domain_id
        role_domain_id = CONF.identity.default_domain_id

        role = PROVIDERS.role_api.create_role(
            uuid.uuid4().hex, unit.new_role_ref(domain_id=role_domain_id)
        )

        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=user_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=project_domain_id)
        )
        with self.test_client() as c:
            c.put(
                '/v3/projects/{}/users/{}/roles/{}'.format(
                    project['id'], user['id'], role['id']
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_create_grant_for_user_other_domain_on_own_domain(self):
        user_domain_id = CONF.identity.default_domain_id
        domain_id = self.domain_id

        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=user_domain_id)
        )

        with self.test_client() as c:
            c.put(
                '/v3/domains/{}/users/{}/roles/{}'.format(
                    domain_id, user['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_create_grant_for_user_own_domain_on_other_domain(self):
        user_domain_id = self.domain_id
        domain_id = CONF.identity.default_domain_id

        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=user_domain_id)
        )

        with self.test_client() as c:
            c.put(
                '/v3/domains/{}/users/{}/roles/{}'.format(
                    domain_id, user['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_create_grant_for_user_own_domain_on_own_domain_with_role_other_domain(
        self,
    ):  # noqa: E501
        user_domain_id = self.domain_id
        domain_id = self.domain_id
        role_domain_id = CONF.identity.default_domain_id

        role = PROVIDERS.role_api.create_role(
            uuid.uuid4().hex, unit.new_role_ref(domain_id=role_domain_id)
        )

        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=user_domain_id)
        )

        with self.test_client() as c:
            c.put(
                '/v3/domains/{}/users/{}/roles/{}'.format(
                    domain_id, user['id'], role['id']
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_create_grant_for_group_other_domain_on_project_own_domain(
        self,
    ):  # noqa: E501
        group_domain_id = CONF.identity.default_domain_id
        project_domain_id = self.domain_id

        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=group_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=project_domain_id)
        )

        with self.test_client() as c:
            c.put(
                '/v3/projects/{}/groups/{}/roles/{}'.format(
                    project['id'],
                    group['id'],
                    self.bootstrapper.reader_role_id,
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_create_grant_for_group_own_domain_on_project_other_domain(
        self,
    ):  # noqa: E501
        group_domain_id = self.domain_id
        project_domain_id = CONF.identity.default_domain_id

        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=group_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=project_domain_id)
        )

        with self.test_client() as c:
            c.put(
                '/v3/projects/{}/groups/{}/roles/{}'.format(
                    project['id'],
                    group['id'],
                    self.bootstrapper.reader_role_id,
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_create_grant_for_group_own_domain_on_project_own_domain_with_role_other_domain(
        self,
    ):  # noqa: E501
        group_domain_id = self.domain_id
        project_domain_id = self.domain_id
        role_domain_id = CONF.identity.default_domain_id

        role = PROVIDERS.role_api.create_role(
            uuid.uuid4().hex, unit.new_role_ref(domain_id=role_domain_id)
        )

        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=group_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=project_domain_id)
        )

        with self.test_client() as c:
            c.put(
                '/v3/projects/{}/groups/{}/roles/{}'.format(
                    project['id'], group['id'], role['id']
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_create_grant_for_group_other_domain_on_own_domain(self):
        group_domain_id = CONF.identity.default_domain_id
        domain_id = self.domain_id

        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=group_domain_id)
        )

        with self.test_client() as c:
            c.put(
                '/v3/domains/{}/groups/{}/roles/{}'.format(
                    domain_id, group['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_create_grant_for_group_own_domain_on_other_domain(self):
        group_domain_id = self.domain_id
        domain_id = CONF.identity.default_domain_id

        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=group_domain_id)
        )

        with self.test_client() as c:
            c.put(
                '/v3/domains/{}/groups/{}/roles/{}'.format(
                    domain_id, group['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_create_grant_for_group_own_domain_on_own_domain_with_role_other_domain(
        self,
    ):  # noqa: E501
        group_domain_id = self.domain_id
        domain_id = self.domain_id
        role_domain_id = CONF.identity.default_domain_id

        role = PROVIDERS.role_api.create_role(
            uuid.uuid4().hex, unit.new_role_ref(domain_id=role_domain_id)
        )

        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=group_domain_id)
        )

        with self.test_client() as c:
            c.put(
                '/v3/domains/{}/groups/{}/roles/{}'.format(
                    domain_id, group['id'], role['id']
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_revoke_grant_from_user_other_domain_on_project_own_domain(
        self,
    ):  # noqa: E501
        user_domain_id = CONF.identity.default_domain_id
        project_domain_id = self.domain_id

        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=user_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=project_domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=user['id'],
            project_id=project['id'],
        )

        with self.test_client() as c:
            c.delete(
                '/v3/projects/{}/users/{}/roles/{}'.format(
                    project['id'], user['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_revoke_grant_from_user_own_domain_on_project_other_domain(
        self,
    ):  # noqa: E501
        user_domain_id = self.domain_id
        project_domain_id = CONF.identity.default_domain_id

        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=user_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=project_domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=user['id'],
            project_id=project['id'],
        )

        with self.test_client() as c:
            c.delete(
                '/v3/projects/{}/users/{}/roles/{}'.format(
                    project['id'], user['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_revoke_grant_from_user_other_domain_on_own_domain(self):
        user_domain_id = CONF.identity.default_domain_id
        domain_id = self.domain_id

        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=user_domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=user['id'],
            domain_id=domain_id,
        )

        with self.test_client() as c:
            c.delete(
                '/v3/domains/{}/users/{}/roles/{}'.format(
                    domain_id, user['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_revoke_grant_from_user_own_domain_on_other_domain(self):
        user_domain_id = self.domain_id
        domain_id = CONF.identity.default_domain_id

        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=user_domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=user['id'],
            domain_id=domain_id,
        )

        with self.test_client() as c:
            c.delete(
                '/v3/domains/{}/users/{}/roles/{}'.format(
                    domain_id, user['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_revoke_grant_from_user_own_domain_on_own_domain_with_role_other_domain(
        self,
    ):  # noqa: E501
        user_domain_id = self.domain_id
        domain_id = self.domain_id
        role_domain_id = CONF.identity.default_domain_id

        role = PROVIDERS.role_api.create_role(
            uuid.uuid4().hex, unit.new_role_ref(domain_id=role_domain_id)
        )

        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=user_domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            role['id'], user_id=user['id'], domain_id=domain_id
        )

        with self.test_client() as c:
            c.delete(
                '/v3/domains/{}/users/{}/roles/{}'.format(
                    domain_id, user['id'], role['id']
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_revoke_grant_from_group_other_domain_on_project_own_domain(
        self,
    ):  # noqa: E501
        group_domain_id = CONF.identity.default_domain_id
        project_domain_id = self.domain_id

        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=group_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=project_domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            group_id=group['id'],
            project_id=project['id'],
        )

        with self.test_client() as c:
            c.delete(
                '/v3/projects/{}/groups/{}/roles/{}'.format(
                    project['id'],
                    group['id'],
                    self.bootstrapper.reader_role_id,
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_revoke_grant_from_group_own_domain_on_project_other_domain(
        self,
    ):  # noqa: E501
        group_domain_id = self.domain_id
        project_domain_id = CONF.identity.default_domain_id

        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=group_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=project_domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            group_id=group['id'],
            project_id=project['id'],
        )

        with self.test_client() as c:
            c.delete(
                '/v3/projects/{}/groups/{}/roles/{}'.format(
                    project['id'],
                    group['id'],
                    self.bootstrapper.reader_role_id,
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_revoke_grant_from_group_other_domain_on_own_domain(self):
        group_domain_id = CONF.identity.default_domain_id
        domain_id = self.domain_id

        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=group_domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            group_id=group['id'],
            domain_id=domain_id,
        )

        with self.test_client() as c:
            c.delete(
                '/v3/domains/{}/groups/{}/roles/{}'.format(
                    domain_id, group['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_revoke_grant_from_group_own_domain_on_other_domain(self):
        group_domain_id = self.domain_id
        domain_id = CONF.identity.default_domain_id

        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=group_domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            group_id=group['id'],
            domain_id=domain_id,
        )

        with self.test_client() as c:
            c.delete(
                '/v3/domains/{}/groups/{}/roles/{}'.format(
                    domain_id, group['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_revoke_grant_from_group_own_domain_on_own_domain_with_role_other_domain(
        self,
    ):  # noqa: E501
        group_domain_id = self.domain_id
        domain_id = self.domain_id
        role_domain_id = CONF.identity.default_domain_id

        role = PROVIDERS.role_api.create_role(
            uuid.uuid4().hex, unit.new_role_ref(domain_id=role_domain_id)
        )

        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=group_domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            role['id'], group_id=group['id'], domain_id=domain_id
        )

        with self.test_client() as c:
            c.delete(
                '/v3/domains/{}/groups/{}/roles/{}'.format(
                    domain_id, group['id'], role['id']
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )


class SystemReaderTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _SystemUserGrantTests,
    _SystemMemberAndReaderGrantTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        system_reader = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(system_reader)['id']
        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.bootstrapper.reader_role_id
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=system_reader['password'],
            system=True,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class SystemMemberTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _SystemUserGrantTests,
    _SystemMemberAndReaderGrantTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        system_member = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(system_member)['id']
        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.bootstrapper.member_role_id
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=system_member['password'],
            system=True,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class SystemAdminTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _SystemUserGrantTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        self.user_id = self.bootstrapper.admin_user_id
        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=self.bootstrapper.admin_password,
            system=True,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def test_can_create_grant_for_user_on_project(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        with self.test_client() as c:
            c.put(
                '/v3/projects/{}/users/{}/roles/{}'.format(
                    project['id'], user['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
            )

    def test_can_create_grant_for_user_on_domain(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        )

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        with self.test_client() as c:
            c.put(
                '/v3/domains/{}/users/{}/roles/{}'.format(
                    domain['id'], user['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
            )

    def test_can_create_grant_for_group_on_project(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=CONF.identity.default_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        with self.test_client() as c:
            c.put(
                '/v3/projects/{}/groups/{}/roles/{}'.format(
                    project['id'],
                    group['id'],
                    self.bootstrapper.reader_role_id,
                ),
                headers=self.headers,
            )

    def test_can_create_grant_for_group_on_domain(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=CONF.identity.default_domain_id)
        )

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        with self.test_client() as c:
            c.put(
                '/v3/domains/{}/groups/{}/roles/{}'.format(
                    domain['id'], group['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
            )

    def test_can_revoke_grant_from_user_on_project(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=user['id'],
            project_id=project['id'],
        )

        with self.test_client() as c:
            c.delete(
                '/v3/projects/{}/users/{}/roles/{}'.format(
                    project['id'], user['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
            )

    def test_can_revoke_grant_from_user_on_domain(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        )

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=user['id'],
            domain_id=domain['id'],
        )

        with self.test_client() as c:
            c.delete(
                '/v3/domains/{}/users/{}/roles/{}'.format(
                    domain['id'], user['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
            )

    def test_can_revoke_grant_from_group_on_project(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=CONF.identity.default_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            group_id=group['id'],
            project_id=project['id'],
        )

        with self.test_client() as c:
            c.delete(
                '/v3/projects/{}/groups/{}/roles/{}'.format(
                    project['id'],
                    group['id'],
                    self.bootstrapper.reader_role_id,
                ),
                headers=self.headers,
            )

    def test_can_revoke_grant_from_group_on_domain(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=CONF.identity.default_domain_id)
        )

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            group_id=group['id'],
            domain_id=domain['id'],
        )

        with self.test_client() as c:
            c.delete(
                '/v3/domains/{}/groups/{}/roles/{}'.format(
                    domain['id'], group['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
            )


class _DomainMemberAndReaderTests:
    def test_cannot_create_grant_for_user_on_project(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        with self.test_client() as c:
            c.put(
                '/v3/projects/{}/users/{}/roles/{}'.format(
                    project['id'], user['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_create_grant_for_user_on_domain(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        with self.test_client() as c:
            c.put(
                '/v3/domains/{}/users/{}/roles/{}'.format(
                    domain['id'], user['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_create_grant_for_group_on_project(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=self.domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=self.domain_id)
        )

        with self.test_client() as c:
            c.put(
                '/v3/projects/{}/groups/{}/roles/{}'.format(
                    project['id'],
                    group['id'],
                    self.bootstrapper.reader_role_id,
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_create_grant_for_group_on_domain(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=self.domain_id)
        )

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        with self.test_client() as c:
            c.put(
                '/v3/domains/{}/groups/{}/roles/{}'.format(
                    domain['id'], group['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_revoke_grant_from_user_on_project(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=self.domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=user['id'],
            project_id=project['id'],
        )

        with self.test_client() as c:
            c.delete(
                '/v3/projects/{}/users/{}/roles/{}'.format(
                    project['id'], user['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_revoke_grant_from_user_on_domain(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=user['id'],
            domain_id=domain['id'],
        )

        with self.test_client() as c:
            c.delete(
                '/v3/domains/{}/users/{}/roles/{}'.format(
                    domain['id'], user['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_revoke_grant_from_group_on_project(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=self.domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            group_id=group['id'],
            project_id=project['id'],
        )

        with self.test_client() as c:
            c.delete(
                '/v3/projects/{}/groups/{}/roles/{}'.format(
                    project['id'],
                    group['id'],
                    self.bootstrapper.reader_role_id,
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_cannot_revoke_grant_from_group_on_domain(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=self.domain_id)
        )

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            group_id=group['id'],
            domain_id=domain['id'],
        )

        with self.test_client() as c:
            c.delete(
                '/v3/domains/{}/groups/{}/roles/{}'.format(
                    domain['id'], group['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )


class DomainReaderTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _DomainUserTests,
    _DomainMemberAndReaderTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        self.domain_id = domain['id']
        domain_user = unit.new_user_ref(domain_id=self.domain_id)
        self.user_id = PROVIDERS.identity_api.create_user(domain_user)['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=self.user_id,
            domain_id=self.domain_id,
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=domain_user['password'],
            domain_id=self.domain_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class DomainMemberTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _DomainUserTests,
    _DomainMemberAndReaderTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        self.domain_id = domain['id']
        domain_user = unit.new_user_ref(domain_id=self.domain_id)
        self.user_id = PROVIDERS.identity_api.create_user(domain_user)['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id,
            user_id=self.user_id,
            domain_id=self.domain_id,
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=domain_user['password'],
            domain_id=self.domain_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class DomainAdminTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _DomainUserTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.policy_file = self.useFixture(temporaryfile.SecureTempFile())
        self.policy_file_name = self.policy_file.file_name
        self.useFixture(
            ksfixtures.Policy(
                self.config_fixture, policy_file=self.policy_file_name
            )
        )
        self._override_policy()
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        self.domain_id = domain['id']
        domain_admin = unit.new_user_ref(domain_id=self.domain_id)
        self.user_id = PROVIDERS.identity_api.create_user(domain_admin)['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.admin_role_id,
            user_id=self.user_id,
            domain_id=self.domain_id,
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=domain_admin['password'],
            domain_id=self.domain_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def _override_policy(self):
        # TODO(lbragstad): Remove this once the deprecated policies in
        # keystone.common.policies.grant have been removed. This is only
        # here to make sure we test the new policies instead of the deprecated
        # ones. Oslo.policy will OR deprecated policies with new policies to
        # maintain compatibility and give operators a chance to update
        # permissions or update policies without breaking users. This will
        # cause these specific tests to fail since we're trying to correct this
        # broken behavior with better scope checking.
        with open(self.policy_file_name, 'w') as f:
            overridden_policies = {
                'identity:list_grants': gp.SYSTEM_READER_OR_DOMAIN_READER_LIST,
                'identity:check_grant': gp.SYSTEM_READER_OR_DOMAIN_READER,
                'identity:create_grant': gp.SYSTEM_ADMIN_OR_DOMAIN_ADMIN,
                'identity:revoke_grant': gp.SYSTEM_ADMIN_OR_DOMAIN_ADMIN,
            }
            f.write(jsonutils.dumps(overridden_policies))

    def test_can_create_grant_for_user_on_project(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=self.domain_id)
        )

        with self.test_client() as c:
            c.put(
                '/v3/projects/{}/users/{}/roles/{}'.format(
                    project['id'], user['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
            )

    def test_can_create_grant_for_user_own_domain_on_own_domain(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )

        with self.test_client() as c:
            c.put(
                '/v3/domains/{}/users/{}/roles/{}'.format(
                    self.domain_id,
                    user['id'],
                    self.bootstrapper.reader_role_id,
                ),
                headers=self.headers,
            )

    def test_can_create_grant_for_group_on_project(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=self.domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=self.domain_id)
        )

        with self.test_client() as c:
            c.put(
                '/v3/projects/{}/groups/{}/roles/{}'.format(
                    project['id'],
                    group['id'],
                    self.bootstrapper.reader_role_id,
                ),
                headers=self.headers,
            )

    def test_can_create_grant_for_group_own_domain_on_own_domain(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=self.domain_id)
        )

        with self.test_client() as c:
            c.put(
                '/v3/domains/{}/groups/{}/roles/{}'.format(
                    self.domain_id,
                    group['id'],
                    self.bootstrapper.reader_role_id,
                ),
                headers=self.headers,
            )

    def test_can_revoke_grant_from_user_on_project(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=self.domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=user['id'],
            project_id=project['id'],
        )

        with self.test_client() as c:
            c.delete(
                '/v3/projects/{}/users/{}/roles/{}'.format(
                    project['id'], user['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
            )

    def test_can_revoke_grant_from_group_on_project(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=self.domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=self.domain_id)
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            group_id=group['id'],
            project_id=project['id'],
        )

        with self.test_client() as c:
            c.delete(
                '/v3/projects/{}/groups/{}/roles/{}'.format(
                    project['id'],
                    group['id'],
                    self.bootstrapper.reader_role_id,
                ),
                headers=self.headers,
            )

    def test_cannot_revoke_grant_from_group_on_domain(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=CONF.identity.default_domain_id)
        )

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            group_id=group['id'],
            domain_id=domain['id'],
        )

        with self.test_client() as c:
            c.delete(
                '/v3/domains/{}/groups/{}/roles/{}'.format(
                    domain['id'], group['id'], self.bootstrapper.reader_role_id
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )



# Copyright 2017 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools

from oslo_log import log

from keystone.common import password_hashers
from keystone.common.password_hashers import bcrypt
from keystone.common.password_hashers import pbkdf2
from keystone.common.password_hashers import scrypt
import keystone.conf
from keystone import exception
from keystone.i18n import _

CONF = keystone.conf.CONF
LOG = log.getLogger(__name__)

SUPPORTED_HASHERS: frozenset[type[password_hashers.PasswordHasher]] = (
    frozenset(
        [scrypt.Scrypt, bcrypt.Bcrypt, bcrypt.Bcrypt_sha256, pbkdf2.Sha512]
    )
)

_HASHER_NAME_MAP = {hasher.name: hasher for hasher in SUPPORTED_HASHERS}


# NOTE(notmorgan): Build the list of prefixes. This comprehension builds
# a dictionary where the keys are the prefix (all hashedpasswords are
# '$<ident>$<metadata>$<hash>') so we can do a fast-lookup on the hasher to
# use. If has hasher has multiple ident options it is encoded in the
# .ident_values attribute whereas hashers that have a single option
# ( ) only has the .ident attribute.
# NOTE(noonedeadpunk): Though bcrypt_sha256 does define <ident> as part of
# the metadata, actual indent is represented with a <prefix> instead.
def _get_hash_ident(hashers):
    for hasher in hashers:
        if hasattr(hasher, 'prefix'):
            ident = (getattr(hasher, 'prefix'),)
        elif hasattr(hasher, 'ident_values'):
            ident = getattr(hasher, 'ident_values')
        else:
            ident = (getattr(hasher, 'ident'),)
        yield (hasher, ident)


_HASHER_IDENT_MAP = {
    prefix: module
    for module, prefix in itertools.chain(
        *[
            zip([mod] * len(ident), ident)
            for mod, ident in _get_hash_ident(SUPPORTED_HASHERS)
        ]
    )
}


def _get_hasher_from_ident(hashed):
    try:
        return _HASHER_IDENT_MAP[hashed[0 : hashed.index('$', 1) + 1]]
    except KeyError:
        raise ValueError(
            _('Unsupported password hashing algorithm ident: %s')
            % hashed[0 : hashed.index('$', 1) + 1]
        )


def verify_length_and_trunc_password(password) -> bytes:
    """Verify and truncate the provided password to the max_password_length.

    We also need to check that the configured password hashing algorithm does
    not silently truncate the password.  For example, passlib.hash.bcrypt does
    this:
    https://passlib.readthedocs.io/en/stable/lib/passlib.hash.bcrypt.html#security-issues

    """
    # When using bcrypt, we limit the password length to 54 to ensure all
    # bytes are fully mixed. See:
    # https://passlib.readthedocs.io/en/stable/lib/passlib.hash.bcrypt.html#security-issues
    BCRYPT_MAX_LENGTH = 72
    if (
        CONF.identity.password_hash_algorithm == 'bcrypt'  # nosec: B105
        and CONF.identity.max_password_length > BCRYPT_MAX_LENGTH
    ):
        msg = (
            "Truncating password to algorithm specific maximum length %d"
            " characters."
        )
        LOG.warning(msg, BCRYPT_MAX_LENGTH)
        max_length = BCRYPT_MAX_LENGTH
    else:
        max_length = CONF.identity.max_password_length

    try:
        password_utf8 = password.encode('utf-8')
        if len(password_utf8) > max_length:
            if CONF.strict_password_check:
                raise exception.PasswordVerificationError(size=max_length)
            else:
                msg = "Truncating user password to %d characters."
                LOG.warning(msg, max_length)
                return password_utf8[:max_length]
        else:
            return password_utf8
    except AttributeError:
        raise exception.ValidationError(attribute='string', target='password')


def check_password(password: str, hashed: str) -> bool:
    """Check that a plaintext password matches hashed.

    hashpw returns the salt value concatenated with the actual hash value.
    It extracts the actual salt if this value is then passed as the salt.

    """
    if password is None or hashed is None:
        return False
    password_utf8 = verify_length_and_trunc_password(password)
    hasher = _get_hasher_from_ident(hashed)
    return hasher.verify(password_utf8, hashed)


def hash_user_password(user):
    """Hash a user dict's password without modifying the passed-in dict."""
    password = user.get('password')
    if password is None:
        return user

    return dict(user, password=hash_password(password))


def hash_password(password: str) -> str:
    """Hash a password. Harder."""
    params = {}
    password_utf8 = verify_length_and_trunc_password(password)
    conf_hasher = CONF.identity.password_hash_algorithm
    hasher = _HASHER_NAME_MAP.get(conf_hasher)

    if hasher is None:
        raise RuntimeError(
            _('Password Hash Algorithm %s not found')
            % CONF.identity.password_hash_algorithm
        )

    if CONF.identity.password_hash_rounds:
        params['rounds'] = CONF.identity.password_hash_rounds
    if hasher is scrypt.Scrypt:
        params["n"] = 16
        if CONF.identity.scrypt_block_size:
            params["r"] = CONF.identity.scrypt_block_size
        if CONF.identity.scrypt_parallelism:
            params["p"] = CONF.identity.scrypt_parallelism
        if CONF.identity.salt_bytesize:
            params["salt_size"] = CONF.identity.salt_bytesize
        return scrypt.Scrypt.hash(password_utf8, **params)
    elif hasher is bcrypt.Bcrypt:
        return bcrypt.Bcrypt.hash(password_utf8, **params)
    elif hasher is bcrypt.Bcrypt_sha256:
        return bcrypt.Bcrypt_sha256.hash(password_utf8, **params)
    elif hasher is pbkdf2.Sha512:
        if CONF.identity.salt_bytesize:
            params["salt_size"] = CONF.identity.salt_bytesize
        return pbkdf2.Sha512.hash(password_utf8, **params)
    else:
        raise RuntimeError(
            _('Password Hash Algorithm %s not implemented')
            % CONF.identity.password_hash_algorithm
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import base64
import datetime
import hashlib
import os
from unittest import mock
import uuid

from oslo_utils import timeutils

from keystone.common import fernet_utils
from keystone.common import provider_api
from keystone.common import utils
import keystone.conf
from keystone import exception
from keystone.identity.backends import resource_options as ro
from keystone.receipt.providers import fernet
from keystone.receipt import receipt_formatters
from keystone.tests import unit
from keystone.tests.unit import default_fixtures
from keystone.tests.unit import ksfixtures
from keystone.tests.unit.ksfixtures import database
from keystone.token import provider as token_provider

CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


class TestFernetReceiptProvider(unit.TestCase):
    def setUp(self):
        super().setUp()
        self.provider = fernet.Provider()

    def test_invalid_receipt_raises_receipt_not_found(self):
        receipt_id = uuid.uuid4().hex
        e = self.assertRaises(
            exception.ReceiptNotFound,
            self.provider.validate_receipt,
            receipt_id,
        )
        self.assertIn(receipt_id, f'{e}')


class TestValidate(unit.TestCase):
    def setUp(self):
        super().setUp()
        self.useFixture(database.Database())
        self.useFixture(
            ksfixtures.ConfigAuthPlugins(
                self.config_fixture, ['totp', 'token', 'password']
            )
        )
        self.load_backends()
        PROVIDERS.resource_api.create_domain(
            default_fixtures.ROOT_DOMAIN['id'], default_fixtures.ROOT_DOMAIN
        )

    def config_overrides(self):
        super().config_overrides()
        self.config_fixture.config(group='receipt', provider='fernet')

    def test_validate_v3_receipt_simple(self):
        # Check the fields in the receipt result when use validate_v3_receipt
        # with a simple receipt.

        domain_ref = unit.new_domain_ref()
        domain_ref = PROVIDERS.resource_api.create_domain(
            domain_ref['id'], domain_ref
        )

        rule_list = [['password', 'totp'], ['password', 'totp', 'token']]

        user_ref = unit.new_user_ref(domain_ref['id'])
        user_ref = PROVIDERS.identity_api.create_user(user_ref)
        user_ref['options'][ro.MFA_RULES_OPT.option_name] = rule_list
        user_ref['options'][ro.MFA_ENABLED_OPT.option_name] = True
        PROVIDERS.identity_api.update_user(user_ref['id'], user_ref)

        method_names = ['password']
        receipt = PROVIDERS.receipt_provider_api.issue_receipt(
            user_ref['id'], method_names
        )

        receipt = PROVIDERS.receipt_provider_api.validate_receipt(receipt.id)
        self.assertIsInstance(receipt.expires_at, str)
        self.assertIsInstance(receipt.issued_at, str)
        self.assertEqual(set(method_names), set(receipt.methods))
        self.assertEqual(
            {frozenset(r) for r in rule_list},
            {frozenset(r) for r in receipt.required_methods},
        )
        self.assertEqual(user_ref['id'], receipt.user_id)

    def test_validate_v3_receipt_validation_error_exc(self):
        # When the receipt format isn't recognized, ReceiptNotFound is raised.

        # A uuid string isn't a valid Fernet receipt.
        receipt_id = uuid.uuid4().hex
        self.assertRaises(
            exception.ReceiptNotFound,
            PROVIDERS.receipt_provider_api.validate_receipt,
            receipt_id,
        )


class TestReceiptFormatter(unit.TestCase):
    def test_restore_padding(self):
        # 'a' will result in '==' padding, 'aa' will result in '=' padding, and
        # 'aaa' will result in no padding.
        binary_to_test = [b'a', b'aa', b'aaa']

        for binary in binary_to_test:
            # base64.urlsafe_b64encode takes bytes and returns
            # bytes.
            encoded_string = base64.urlsafe_b64encode(binary)
            encoded_string = encoded_string.decode('utf-8')
            # encoded_string is now str.
            encoded_str_without_padding = encoded_string.rstrip('=')
            self.assertFalse(encoded_str_without_padding.endswith('='))
            encoded_str_with_padding_restored = (
                receipt_formatters.ReceiptFormatter.restore_padding(
                    encoded_str_without_padding
                )
            )
            self.assertEqual(encoded_string, encoded_str_with_padding_restored)


class TestPayloads(unit.TestCase):
    def setUp(self):
        super().setUp()
        self.useFixture(
            ksfixtures.ConfigAuthPlugins(
                self.config_fixture, ['totp', 'token', 'password']
            )
        )

    def assertTimestampsEqual(self, expected, actual):
        # The timestamp that we get back when parsing the payload may not
        # exactly match the timestamp that was put in the payload due to
        # conversion to and from a float.

        exp_time = timeutils.parse_isotime(expected)
        actual_time = timeutils.parse_isotime(actual)

        # the granularity of timestamp string is microseconds and it's only the
        # last digit in the representation that's different, so use a delta
        # just above nanoseconds.
        return self.assertCloseEnoughForGovernmentWork(
            exp_time, actual_time, delta=1e-05
        )

    def test_strings_can_be_converted_to_bytes(self):
        s = token_provider.random_urlsafe_str()
        self.assertIsInstance(s, str)

        b = receipt_formatters.ReceiptPayload.random_urlsafe_str_to_bytes(s)
        self.assertIsInstance(b, bytes)

    def test_uuid_hex_to_byte_conversions(self):
        payload_cls = receipt_formatters.ReceiptPayload

        expected_hex_uuid = uuid.uuid4().hex
        uuid_obj = uuid.UUID(expected_hex_uuid)
        expected_uuid_in_bytes = uuid_obj.bytes
        actual_uuid_in_bytes = payload_cls.convert_uuid_hex_to_bytes(
            expected_hex_uuid
        )
        self.assertEqual(expected_uuid_in_bytes, actual_uuid_in_bytes)
        actual_hex_uuid = payload_cls.convert_uuid_bytes_to_hex(
            expected_uuid_in_bytes
        )
        self.assertEqual(expected_hex_uuid, actual_hex_uuid)

    def test_time_string_to_float_conversions(self):
        payload_cls = receipt_formatters.ReceiptPayload

        original_time_str = utils.isotime(subsecond=True)
        time_obj = timeutils.parse_isotime(original_time_str)
        expected_time_float = (
            timeutils.normalize_time(time_obj)
            - datetime.datetime.fromtimestamp(
                0, datetime.timezone.utc
            ).replace(tzinfo=None)
        ).total_seconds()

        # NOTE(lbragstad): The receipt expiration time for Fernet receipts is
        # passed in the payload of the receipt. This is different from the
        # receipt creation time, which is handled by Fernet and doesn't support
        # subsecond precision because it is a timestamp integer.
        self.assertIsInstance(expected_time_float, float)

        actual_time_float = payload_cls._convert_time_string_to_float(
            original_time_str
        )
        self.assertIsInstance(actual_time_float, float)
        self.assertEqual(expected_time_float, actual_time_float)

        # Generate expected_time_str using the same time float. Using
        # original_time_str from utils.isotime will occasionally fail due to
        # floating point rounding differences.
        time_object = datetime.datetime.fromtimestamp(
            actual_time_float, datetime.timezone.utc
        ).replace(tzinfo=None)
        expected_time_str = utils.isotime(time_object, subsecond=True)

        actual_time_str = payload_cls._convert_float_to_time_string(
            actual_time_float
        )
        self.assertEqual(expected_time_str, actual_time_str)

    def _test_payload(self, payload_class, exp_user_id=None, exp_methods=None):
        exp_user_id = exp_user_id or uuid.uuid4().hex
        exp_methods = exp_methods or ['password']
        exp_expires_at = utils.isotime(timeutils.utcnow(), subsecond=True)

        payload = payload_class.assemble(
            exp_user_id, exp_methods, exp_expires_at
        )

        (user_id, methods, expires_at) = payload_class.disassemble(payload)

        self.assertEqual(exp_user_id, user_id)
        self.assertEqual(exp_methods, methods)
        self.assertTimestampsEqual(exp_expires_at, expires_at)

    def test_payload(self):
        self._test_payload(receipt_formatters.ReceiptPayload)

    def test_payload_multiple_methods(self):
        self._test_payload(
            receipt_formatters.ReceiptPayload, exp_methods=['password', 'totp']
        )


class TestFernetKeyRotation(unit.TestCase):
    def setUp(self):
        super().setUp()

        # A collection of all previously-seen signatures of the key
        # repository's contents.
        self.key_repo_signatures = set()

    @property
    def keys(self):
        """Key files converted to numbers."""
        return sorted(
            int(x) for x in os.listdir(CONF.fernet_receipts.key_repository)
        )

    @property
    def key_repository_size(self):
        """The number of keys in the key repository."""
        return len(self.keys)

    @property
    def key_repository_signature(self):
        """Create a "thumbprint" of the current key repository.

        Because key files are renamed, this produces a hash of the contents of
        the key files, ignoring their filenames.

        The resulting signature can be used, for example, to ensure that you
        have a unique set of keys after you perform a key rotation (taking a
        static set of keys, and simply shuffling them, would fail such a test).

        """
        # Load the keys into a list, keys is list of str.
        key_utils = fernet_utils.FernetUtils(
            CONF.fernet_receipts.key_repository,
            CONF.fernet_receipts.max_active_keys,
            'fernet_receipts',
        )
        keys = key_utils.load_keys()

        # Sort the list of keys by the keys themselves (they were previously
        # sorted by filename).
        keys.sort()

        # Create the thumbprint using all keys in the repository.
        signature = hashlib.sha1()
        for key in keys:
            # Need to convert key to bytes for update.
            signature.update(key.encode('utf-8'))
        return signature.hexdigest()

    def assertRepositoryState(self, expected_size):
        """Validate the state of the key repository."""
        self.assertEqual(expected_size, self.key_repository_size)
        self.assertUniqueRepositoryState()

    def assertUniqueRepositoryState(self):
        """Ensure that the current key repo state has not been seen before."""
        # This is assigned to a variable because it takes some work to
        # calculate.
        signature = self.key_repository_signature

        # Ensure the signature is not in the set of previously seen signatures.
        self.assertNotIn(signature, self.key_repo_signatures)

        # Add the signature to the set of repository signatures to validate
        # that we don't see it again later.
        self.key_repo_signatures.add(signature)

    def test_rotation(self):
        # Initializing a key repository results in this many keys. We don't
        # support max_active_keys being set any lower.
        min_active_keys = 2

        # Simulate every rotation strategy up to "rotating once a week while
        # maintaining a year's worth of keys."
        for max_active_keys in range(min_active_keys, 52 + 1):
            self.config_fixture.config(
                group='fernet_receipts', max_active_keys=max_active_keys
            )

            # Ensure that resetting the key repository always results in 2
            # active keys.
            self.useFixture(
                ksfixtures.KeyRepository(
                    self.config_fixture,
                    'fernet_receipts',
                    CONF.fernet_receipts.max_active_keys,
                )
            )

            # Validate the initial repository state.
            self.assertRepositoryState(expected_size=min_active_keys)

            # The repository should be initialized with a staged key (0) and a
            # primary key (1). The next key is just auto-incremented.
            exp_keys = [0, 1]
            next_key_number = exp_keys[-1] + 1  # keep track of next key
            self.assertEqual(exp_keys, self.keys)

            # Rotate the keys just enough times to fully populate the key
            # repository.
            key_utils = fernet_utils.FernetUtils(
                CONF.fernet_receipts.key_repository,
                CONF.fernet_receipts.max_active_keys,
                'fernet_receipts',
            )
            for rotation in range(max_active_keys - min_active_keys):
                key_utils.rotate_keys()
                self.assertRepositoryState(expected_size=rotation + 3)

                exp_keys.append(next_key_number)
                next_key_number += 1
                self.assertEqual(exp_keys, self.keys)

            # We should have a fully populated key repository now.
            self.assertEqual(max_active_keys, self.key_repository_size)

            # Rotate an additional number of times to ensure that we maintain
            # the desired number of active keys.
            key_utils = fernet_utils.FernetUtils(
                CONF.fernet_receipts.key_repository,
                CONF.fernet_receipts.max_active_keys,
                'fernet_receipts',
            )
            for rotation in range(10):
                key_utils.rotate_keys()
                self.assertRepositoryState(expected_size=max_active_keys)

                exp_keys.pop(1)
                exp_keys.append(next_key_number)
                next_key_number += 1
                self.assertEqual(exp_keys, self.keys)

    def test_rotation_disk_write_fail(self):
        # Make sure that the init key repository contains 2 keys
        self.assertRepositoryState(expected_size=2)

        key_utils = fernet_utils.FernetUtils(
            CONF.fernet_receipts.key_repository,
            CONF.fernet_receipts.max_active_keys,
            'fernet_receipts',
        )

        # Simulate the disk full situation
        mock_open = mock.mock_open()
        file_handle = mock_open()
        file_handle.flush.side_effect = OSError('disk full')

        with mock.patch('keystone.common.fernet_utils.open', mock_open):
            self.assertRaises(IOError, key_utils.rotate_keys)

        # Assert that the key repository is unchanged
        self.assertEqual(self.key_repository_size, 2)

        with mock.patch('keystone.common.fernet_utils.open', mock_open):
            self.assertRaises(IOError, key_utils.rotate_keys)

        # Assert that the key repository is still unchanged, even after
        # repeated rotation attempts.
        self.assertEqual(self.key_repository_size, 2)

        # Rotate the keys normally, without any mocking, to show that the
        # system can recover.
        key_utils.rotate_keys()

        # Assert that the key repository is now expanded.
        self.assertEqual(self.key_repository_size, 3)

    def test_rotation_empty_file(self):
        active_keys = 2
        self.assertRepositoryState(expected_size=active_keys)
        empty_file = os.path.join(CONF.fernet_receipts.key_repository, '2')
        with open(empty_file, 'w'):
            pass
        key_utils = fernet_utils.FernetUtils(
            CONF.fernet_receipts.key_repository,
            CONF.fernet_receipts.max_active_keys,
            'fernet_receipts',
        )
        # Rotate the keys to overwrite the empty file
        key_utils.rotate_keys()
        self.assertTrue(os.path.isfile(empty_file))
        keys = key_utils.load_keys()
        self.assertEqual(3, len(keys))
        self.assertTrue(os.path.getsize(empty_file) > 0)

    def test_non_numeric_files(self):
        evil_file = os.path.join(CONF.fernet_receipts.key_repository, '99.bak')
        with open(evil_file, 'w'):
            pass
        key_utils = fernet_utils.FernetUtils(
            CONF.fernet_receipts.key_repository,
            CONF.fernet_receipts.max_active_keys,
            'fernet_receipts',
        )
        key_utils.rotate_keys()
        self.assertTrue(os.path.isfile(evil_file))
        keys = 0
        for x in os.listdir(CONF.fernet_receipts.key_repository):
            if x == '99.bak':
                continue
            keys += 1
        self.assertEqual(3, keys)


class TestLoadKeys(unit.TestCase):
    def assertValidFernetKeys(self, keys):
        # Make sure each key is a non-empty string
        for key in keys:
            self.assertGreater(len(key), 0)
            self.assertIsInstance(key, str)

    def test_non_numeric_files(self):
        evil_file = os.path.join(CONF.fernet_receipts.key_repository, '~1')
        with open(evil_file, 'w'):
            pass
        key_utils = fernet_utils.FernetUtils(
            CONF.fernet_receipts.key_repository,
            CONF.fernet_receipts.max_active_keys,
            'fernet_receipts',
        )
        keys = key_utils.load_keys()
        self.assertEqual(2, len(keys))
        self.assertValidFernetKeys(keys)

    def test_empty_files(self):
        empty_file = os.path.join(CONF.fernet_receipts.key_repository, '2')
        with open(empty_file, 'w'):
            pass
        key_utils = fernet_utils.FernetUtils(
            CONF.fernet_receipts.key_repository,
            CONF.fernet_receipts.max_active_keys,
            'fernet_receipts',
        )
        keys = key_utils.load_keys()
        self.assertEqual(2, len(keys))
        self.assertValidFernetKeys(keys)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystone.application_credential.backends import sql as sql_driver
from keystone.common import provider_api
from keystone.common import sql
from keystone.tests.unit.application_credential import test_backends
from keystone.tests.unit.backend import core_sql
from keystone.tests.unit.ksfixtures import database

PROVIDERS = provider_api.ProviderAPIs


class SQLModelTestCase(core_sql.BaseBackendSqlModels):
    def test_application_credential_model(self):
        cols = (
            ('internal_id', sql.Integer, None),
            ('id', sql.String, 64),
            ('name', sql.String, 255),
            ('secret_hash', sql.String, 255),
            ('description', sql.Text, None),
            ('user_id', sql.String, 64),
            ('project_id', sql.String, 64),
            ('system', sql.String, 64),
            ('expires_at', sql.DateTimeInt, None),
        )
        self.assertExpectedSchema('application_credential', cols)

    def test_application_credential_role_model(self):
        cols = (
            ('application_credential_id', sql.Integer, None),
            ('role_id', sql.String, 64),
        )
        self.assertExpectedSchema('application_credential_role', cols)

    def test_access_rule_model(self):
        cols = (
            ('id', sql.Integer, None),
            ('external_id', sql.String, 64),
            ('user_id', sql.String, 64),
            ('service', sql.String, 64),
            ('path', sql.String, 128),
            ('method', sql.String, 16),
        )
        self.assertExpectedSchema('access_rule', cols)

    def test_application_credential_access_rule_model(self):
        cols = (
            ('application_credential_id', sql.Integer, None),
            ('access_rule_id', sql.Integer, None),
        )
        self.assertExpectedSchema('application_credential_access_rule', cols)


class SQLDriverTestCase(
    core_sql.BaseBackendSqlTests, test_backends.ApplicationCredentialTests
):
    def setUp(self):
        self.useFixture(database.Database())
        self.driver = sql_driver.ApplicationCredential()
        super().setUp()

        self.app_cred_api = PROVIDERS.application_credential_api



# Copyright 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystone.common import sql
from keystone.tests.unit import test_backend_sql


class SqlFederation(test_backend_sql.SqlModels):
    """Set of tests for checking SQL Federation."""

    def test_identity_provider(self):
        cols = (
            ('id', sql.String, 64),
            ('domain_id', sql.String, 64),
            ('enabled', sql.Boolean, None),
            ('description', sql.Text, None),
            ('authorization_ttl', sql.Integer, None),
        )
        self.assertExpectedSchema('identity_provider', cols)

    def test_idp_remote_ids(self):
        cols = (('idp_id', sql.String, 64), ('remote_id', sql.String, 255))
        self.assertExpectedSchema('idp_remote_ids', cols)

    def test_federated_protocol(self):
        cols = (
            ('id', sql.String, 64),
            ('idp_id', sql.String, 64),
            ('mapping_id', sql.String, 64),
            ('remote_id_attribute', sql.String, 64),
        )
        self.assertExpectedSchema('federation_protocol', cols)

    def test_mapping(self):
        cols = (
            ('id', sql.String, 64),
            ('rules', sql.JsonBlob, None),
            ('schema_version', sql.String, 5),
        )
        self.assertExpectedSchema('mapping', cols)

    def test_service_provider(self):
        cols = (
            ('auth_url', sql.String, 256),
            ('id', sql.String, 64),
            ('enabled', sql.Boolean, None),
            ('description', sql.Text, None),
            ('relay_state_prefix', sql.String, 256),
            ('sp_url', sql.String, 256),
        )
        self.assertExpectedSchema('service_provider', cols)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from keystone.common import driver_hints
from keystone.common import provider_api
from keystone import exception
from keystone.tests import unit

PROVIDERS = provider_api.ProviderAPIs


class RegisteredLimitTests:
    def test_create_registered_limit_crud(self):
        # create one, return it.
        registered_limit_1 = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            default_limit=10,
            id=uuid.uuid4().hex,
            description='test description',
        )
        reg_limits = PROVIDERS.unified_limit_api.create_registered_limits(
            [registered_limit_1]
        )
        self.assertDictEqual(registered_limit_1, reg_limits[0])

        # create another two, return them.
        registered_limit_2 = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='snapshot',
            default_limit=5,
            id=uuid.uuid4().hex,
        )
        registered_limit_3 = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='backup',
            default_limit=5,
            id=uuid.uuid4().hex,
        )
        reg_limits = PROVIDERS.unified_limit_api.create_registered_limits(
            [registered_limit_2, registered_limit_3]
        )
        self.assertEqual(2, len(reg_limits))
        for reg_limit in reg_limits:
            if reg_limit['id'] == registered_limit_2['id']:
                self.assertDictEqual(registered_limit_2, reg_limit)
            if reg_limit['id'] == registered_limit_3['id']:
                self.assertDictEqual(registered_limit_3, reg_limit)

    def test_create_registered_limit_duplicate(self):
        registered_limit_1 = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            default_limit=10,
            id=uuid.uuid4().hex,
        )
        PROVIDERS.unified_limit_api.create_registered_limits(
            [registered_limit_1]
        )

        registered_limit_2 = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            default_limit=10,
            id=uuid.uuid4().hex,
        )
        self.assertRaises(
            exception.Conflict,
            PROVIDERS.unified_limit_api.create_registered_limits,
            [registered_limit_2],
        )

    def test_create_multi_registered_limits_duplicate(self):
        registered_limit_1 = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            default_limit=10,
            id=uuid.uuid4().hex,
        )
        PROVIDERS.unified_limit_api.create_registered_limits(
            [registered_limit_1]
        )

        # Create with a duplicated one and a normal one. Both of them will not
        # be created.
        registered_limit_2 = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            default_limit=10,
            id=uuid.uuid4().hex,
        )
        registered_limit_3 = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='snapshot',
            default_limit=10,
            id=uuid.uuid4().hex,
        )
        self.assertRaises(
            exception.Conflict,
            PROVIDERS.unified_limit_api.create_registered_limits,
            [registered_limit_2, registered_limit_3],
        )

        reg_limits = PROVIDERS.unified_limit_api.list_registered_limits()
        self.assertEqual(1, len(reg_limits))
        self.assertEqual(registered_limit_1['id'], reg_limits[0]['id'])

    def test_create_registered_limit_invalid_service(self):
        registered_limit_1 = unit.new_registered_limit_ref(
            service_id=uuid.uuid4().hex,
            region_id=self.region_one['id'],
            resource_name='volume',
            default_limit=10,
            id=uuid.uuid4().hex,
        )
        self.assertRaises(
            exception.ValidationError,
            PROVIDERS.unified_limit_api.create_registered_limits,
            [registered_limit_1],
        )

    def test_create_registered_limit_invalid_region(self):
        registered_limit_1 = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            region_id=uuid.uuid4().hex,
            resource_name='volume',
            default_limit=10,
            id=uuid.uuid4().hex,
        )
        self.assertRaises(
            exception.ValidationError,
            PROVIDERS.unified_limit_api.create_registered_limits,
            [registered_limit_1],
        )

    def test_create_registered_limit_description_none(self):
        registered_limit = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            default_limit=10,
            id=uuid.uuid4().hex,
            description=None,
        )
        res = PROVIDERS.unified_limit_api.create_registered_limits(
            [registered_limit]
        )
        self.assertIsNone(res[0]['description'])

    def test_create_registered_limit_without_description(self):
        registered_limit = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            default_limit=10,
            id=uuid.uuid4().hex,
        )
        registered_limit.pop('description')
        res = PROVIDERS.unified_limit_api.create_registered_limits(
            [registered_limit]
        )
        self.assertIsNone(res[0]['description'])

    def test_update_registered_limit(self):
        # create two registered limits
        registered_limit_1 = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            default_limit=10,
            id=uuid.uuid4().hex,
        )
        registered_limit_2 = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='snapshot',
            default_limit=5,
            id=uuid.uuid4().hex,
        )
        PROVIDERS.unified_limit_api.create_registered_limits(
            [registered_limit_1, registered_limit_2]
        )

        expect_region = 'region_two'
        registered_limit_update = {
            'id': registered_limit_1['id'],
            'region_id': expect_region,
        }
        res = PROVIDERS.unified_limit_api.update_registered_limit(
            registered_limit_1['id'], registered_limit_update
        )
        self.assertEqual(expect_region, res['region_id'])

        # 'id' can be omitted in the update body
        registered_limit_update = {'region_id': expect_region}
        res = PROVIDERS.unified_limit_api.update_registered_limit(
            registered_limit_2['id'], registered_limit_update
        )
        self.assertEqual(expect_region, res['region_id'])

    def test_update_registered_limit_invalid_input_return_bad_request(self):
        registered_limit_1 = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            default_limit=10,
            id=uuid.uuid4().hex,
        )
        PROVIDERS.unified_limit_api.create_registered_limits(
            [registered_limit_1]
        )

        update_ref = {
            'id': registered_limit_1['id'],
            'service_id': uuid.uuid4().hex,
        }
        self.assertRaises(
            exception.ValidationError,
            PROVIDERS.unified_limit_api.update_registered_limit,
            registered_limit_1['id'],
            update_ref,
        )

        update_ref = {'id': registered_limit_1['id'], 'region_id': 'fake_id'}
        self.assertRaises(
            exception.ValidationError,
            PROVIDERS.unified_limit_api.update_registered_limit,
            registered_limit_1['id'],
            update_ref,
        )

    def test_update_registered_limit_duplicate(self):
        registered_limit_1 = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            default_limit=10,
            id=uuid.uuid4().hex,
        )
        registered_limit_2 = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            region_id=self.region_two['id'],
            resource_name='snapshot',
            default_limit=10,
            id=uuid.uuid4().hex,
        )
        PROVIDERS.unified_limit_api.create_registered_limits(
            [registered_limit_1, registered_limit_2]
        )

        # Update registered_limit1 to registered_limit2
        update_ref = {
            'id': registered_limit_1['id'],
            'region_id': self.region_two['id'],
            'resource_name': 'snapshot',
        }
        self.assertRaises(
            exception.Conflict,
            PROVIDERS.unified_limit_api.update_registered_limit,
            registered_limit_1['id'],
            update_ref,
        )

    def test_update_registered_limit_when_reference_limit_exist(self):
        registered_limit_1 = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            default_limit=10,
            id=uuid.uuid4().hex,
        )
        PROVIDERS.unified_limit_api.create_registered_limits(
            [registered_limit_1]
        )
        limit_1 = unit.new_limit_ref(
            project_id=self.project_bar['id'],
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            resource_limit=10,
            id=uuid.uuid4().hex,
        )
        PROVIDERS.unified_limit_api.create_limits([limit_1])

        registered_limit_update = {
            'id': registered_limit_1['id'],
            'region_id': 'region_two',
        }

        self.assertRaises(
            exception.RegisteredLimitError,
            PROVIDERS.unified_limit_api.update_registered_limit,
            registered_limit_1['id'],
            registered_limit_update,
        )

        registered_limit_2 = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            resource_name='volume',
            default_limit=10,
            id=uuid.uuid4().hex,
        )
        PROVIDERS.unified_limit_api.create_registered_limits(
            [registered_limit_2]
        )
        limit_2 = unit.new_limit_ref(
            project_id=self.project_bar['id'],
            service_id=self.service_one['id'],
            resource_name='volume',
            resource_limit=10,
            id=uuid.uuid4().hex,
        )
        PROVIDERS.unified_limit_api.create_limits([limit_2])

        registered_limit_update = {
            'id': registered_limit_2['id'],
            'region_id': 'region_two',
        }

        self.assertRaises(
            exception.RegisteredLimitError,
            PROVIDERS.unified_limit_api.update_registered_limit,
            registered_limit_2['id'],
            registered_limit_update,
        )

    def test_list_registered_limits(self):
        # create two registered limits
        registered_limit_1 = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            default_limit=10,
            id=uuid.uuid4().hex,
        )
        registered_limit_2 = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='snapshot',
            default_limit=5,
            id=uuid.uuid4().hex,
        )
        reg_limits_1 = PROVIDERS.unified_limit_api.create_registered_limits(
            [registered_limit_1, registered_limit_2]
        )

        # list
        reg_limits_2 = PROVIDERS.unified_limit_api.list_registered_limits()
        self.assertEqual(2, len(reg_limits_2))
        self.assertDictEqual(reg_limits_1[0], reg_limits_2[0])
        self.assertDictEqual(reg_limits_1[1], reg_limits_2[1])

    def test_list_registered_limit_by_limit(self):
        self.config_fixture.config(list_limit=1)
        # create two registered limits
        registered_limit_1 = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            default_limit=10,
            id=uuid.uuid4().hex,
        )
        registered_limit_2 = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='snapshot',
            default_limit=5,
            id=uuid.uuid4().hex,
        )
        PROVIDERS.unified_limit_api.create_registered_limits(
            [registered_limit_1, registered_limit_2]
        )

        # list, limit is 1
        hints = driver_hints.Hints()
        reg_limits = PROVIDERS.unified_limit_api.list_registered_limits(
            hints=hints
        )
        self.assertEqual(1, len(reg_limits))

        if reg_limits[0]['id'] == registered_limit_1['id']:
            self.assertDictEqual(registered_limit_1, reg_limits[0])
        else:
            self.assertDictEqual(registered_limit_2, reg_limits[0])

    def test_list_registered_limit_by_filter(self):
        registered_limit_1 = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            default_limit=10,
            id=uuid.uuid4().hex,
        )
        registered_limit_2 = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            region_id=self.region_two['id'],
            resource_name='snapshot',
            default_limit=10,
            id=uuid.uuid4().hex,
        )
        PROVIDERS.unified_limit_api.create_registered_limits(
            [registered_limit_1, registered_limit_2]
        )

        hints = driver_hints.Hints()
        hints.add_filter('service_id', self.service_one['id'])
        res = PROVIDERS.unified_limit_api.list_registered_limits(hints)
        self.assertEqual(2, len(res))

        hints = driver_hints.Hints()
        hints.add_filter('region_id', self.region_one['id'])
        res = PROVIDERS.unified_limit_api.list_registered_limits(hints)
        self.assertEqual(1, len(res))

        hints = driver_hints.Hints()
        hints.add_filter('resource_name', 'backup')
        res = PROVIDERS.unified_limit_api.list_registered_limits(hints)
        self.assertEqual(0, len(res))

    def test_get_registered_limit(self):
        # create two registered limits
        registered_limit_1 = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            default_limit=10,
            id=uuid.uuid4().hex,
        )
        registered_limit_2 = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='snapshot',
            default_limit=5,
            id=uuid.uuid4().hex,
        )
        PROVIDERS.unified_limit_api.create_registered_limits(
            [registered_limit_1, registered_limit_2]
        )

        # show one
        res = PROVIDERS.unified_limit_api.get_registered_limit(
            registered_limit_2['id']
        )
        self.assertDictEqual(registered_limit_2, res)

    def test_get_registered_limit_returns_not_found(self):
        self.assertRaises(
            exception.RegisteredLimitNotFound,
            PROVIDERS.unified_limit_api.get_registered_limit,
            uuid.uuid4().hex,
        )

    def test_delete_registered_limit(self):
        # create two registered limits
        registered_limit_1 = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            default_limit=10,
            id=uuid.uuid4().hex,
        )
        registered_limit_2 = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='snapshot',
            default_limit=5,
            id=uuid.uuid4().hex,
        )
        PROVIDERS.unified_limit_api.create_registered_limits(
            [registered_limit_1, registered_limit_2]
        )

        # delete one
        PROVIDERS.unified_limit_api.delete_registered_limit(
            registered_limit_1['id']
        )
        self.assertRaises(
            exception.RegisteredLimitNotFound,
            PROVIDERS.unified_limit_api.get_registered_limit,
            registered_limit_1['id'],
        )
        reg_limits = PROVIDERS.unified_limit_api.list_registered_limits()
        self.assertEqual(1, len(reg_limits))
        self.assertEqual(registered_limit_2['id'], reg_limits[0]['id'])

    def test_delete_registered_limit_returns_not_found(self):
        self.assertRaises(
            exception.RegisteredLimitNotFound,
            PROVIDERS.unified_limit_api.delete_registered_limit,
            uuid.uuid4().hex,
        )

    def test_delete_registered_limit_when_reference_limit_exist(self):
        registered_limit_1 = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            default_limit=10,
            id=uuid.uuid4().hex,
        )
        PROVIDERS.unified_limit_api.create_registered_limits(
            [registered_limit_1]
        )
        limit_1 = unit.new_limit_ref(
            project_id=self.project_bar['id'],
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            resource_limit=10,
            id=uuid.uuid4().hex,
        )
        PROVIDERS.unified_limit_api.create_limits([limit_1])

        self.assertRaises(
            exception.RegisteredLimitError,
            PROVIDERS.unified_limit_api.delete_registered_limit,
            registered_limit_1['id'],
        )

        registered_limit_2 = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            resource_name='volume',
            default_limit=10,
            id=uuid.uuid4().hex,
        )
        PROVIDERS.unified_limit_api.create_registered_limits(
            [registered_limit_2]
        )
        limit_2 = unit.new_limit_ref(
            project_id=self.project_bar['id'],
            service_id=self.service_one['id'],
            resource_name='volume',
            resource_limit=10,
            id=uuid.uuid4().hex,
        )
        PROVIDERS.unified_limit_api.create_limits([limit_2])

        self.assertRaises(
            exception.RegisteredLimitError,
            PROVIDERS.unified_limit_api.delete_registered_limit,
            registered_limit_2['id'],
        )


class LimitTests:
    def test_default_enforcement_model_is_flat(self):
        expected = {
            'description': (
                'Limit enforcement and validation does not take '
                'project hierarchy into consideration.'
            ),
            'name': 'flat',
        }
        self.assertEqual(expected, PROVIDERS.unified_limit_api.get_model())

    def test_registering_unsupported_enforcement_model_fails(self):
        self.assertRaises(
            ValueError,
            self.config_fixture.config,
            group='unified_limit',
            enforcement_model=uuid.uuid4().hex,
        )

    def test_create_project_limit(self):
        # create one, return it.
        limit_1 = unit.new_limit_ref(
            project_id=self.project_bar['id'],
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            resource_limit=10,
            id=uuid.uuid4().hex,
            description='test description',
            domain_id=None,
        )
        limits = PROVIDERS.unified_limit_api.create_limits([limit_1])
        self.assertDictEqual(limit_1, limits[0])

        # create another two, return them.
        limit_2 = unit.new_limit_ref(
            project_id=self.project_bar['id'],
            service_id=self.service_one['id'],
            region_id=self.region_two['id'],
            resource_name='snapshot',
            resource_limit=5,
            id=uuid.uuid4().hex,
            domain_id=None,
        )
        limit_3 = unit.new_limit_ref(
            project_id=self.project_bar['id'],
            service_id=self.service_one['id'],
            region_id=self.region_two['id'],
            resource_name='backup',
            resource_limit=5,
            id=uuid.uuid4().hex,
            domain_id=None,
        )

        limits = PROVIDERS.unified_limit_api.create_limits([limit_2, limit_3])
        for limit in limits:
            if limit['id'] == limit_2['id']:
                self.assertDictEqual(limit_2, limit)
            if limit['id'] == limit_3['id']:
                self.assertDictEqual(limit_3, limit)

    def test_create_domain_limit(self):
        limit_1 = unit.new_limit_ref(
            project_id=None,
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            resource_limit=10,
            id=uuid.uuid4().hex,
            description='test description',
            domain_id=self.domain_default['id'],
        )
        limits = PROVIDERS.unified_limit_api.create_limits([limit_1])
        self.assertDictEqual(limit_1, limits[0])

    def test_create_project_limit_duplicate(self):
        limit_1 = unit.new_limit_ref(
            project_id=self.project_bar['id'],
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            resource_limit=10,
            id=uuid.uuid4().hex,
        )
        PROVIDERS.unified_limit_api.create_limits([limit_1])

        # use different id but the same project_id, service_id and region_id
        limit_1 = unit.new_limit_ref(
            project_id=self.project_bar['id'],
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            resource_limit=10,
            id=uuid.uuid4().hex,
        )
        self.assertRaises(
            exception.Conflict,
            PROVIDERS.unified_limit_api.create_limits,
            [limit_1],
        )

    def test_create_domain_limit_duplicate(self):
        limit_1 = unit.new_limit_ref(
            project_id=None,
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            resource_limit=10,
            id=uuid.uuid4().hex,
            domain_id=self.domain_default['id'],
        )
        PROVIDERS.unified_limit_api.create_limits([limit_1])

        # use different id but the same domain_id, service_id and region_id
        limit_1 = unit.new_limit_ref(
            project_id=None,
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            resource_limit=10,
            id=uuid.uuid4().hex,
            domain_id=self.domain_default['id'],
        )
        self.assertRaises(
            exception.Conflict,
            PROVIDERS.unified_limit_api.create_limits,
            [limit_1],
        )

    def test_create_limit_with_invalid_service_raises_validation_error(self):
        limit = unit.new_limit_ref(
            project_id=self.project_bar['id'],
            service_id=uuid.uuid4().hex,
            region_id=self.region_one['id'],
            resource_name='volume',
            resource_limit=10,
            id=uuid.uuid4().hex,
        )
        self.assertRaises(
            exception.ValidationError,
            PROVIDERS.unified_limit_api.create_limits,
            [limit],
        )

    def test_create_limit_with_invalid_region_raises_validation_error(self):
        limit = unit.new_limit_ref(
            project_id=self.project_bar['id'],
            service_id=self.service_one['id'],
            region_id=uuid.uuid4().hex,
            resource_name='volume',
            resource_limit=10,
            id=uuid.uuid4().hex,
        )
        self.assertRaises(
            exception.ValidationError,
            PROVIDERS.unified_limit_api.create_limits,
            [limit],
        )

    def test_create_limit_without_reference_registered_limit(self):
        limit_1 = unit.new_limit_ref(
            project_id=self.project_bar['id'],
            service_id=self.service_one['id'],
            region_id=self.region_two['id'],
            resource_name='volume',
            resource_limit=10,
            id=uuid.uuid4().hex,
        )
        self.assertRaises(
            exception.NoLimitReference,
            PROVIDERS.unified_limit_api.create_limits,
            [limit_1],
        )

    def test_create_limit_description_none(self):
        limit = unit.new_limit_ref(
            project_id=self.project_bar['id'],
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            resource_limit=10,
            id=uuid.uuid4().hex,
            description=None,
        )
        res = PROVIDERS.unified_limit_api.create_limits([limit])
        self.assertIsNone(res[0]['description'])

    def test_create_limit_without_description(self):
        limit = unit.new_limit_ref(
            project_id=self.project_bar['id'],
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            resource_limit=10,
            id=uuid.uuid4().hex,
        )
        limit.pop('description')
        res = PROVIDERS.unified_limit_api.create_limits([limit])
        self.assertIsNone(res[0]['description'])

    def test_update_limit(self):
        # create two limits
        limit_1 = unit.new_limit_ref(
            project_id=self.project_bar['id'],
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            resource_limit=10,
            id=uuid.uuid4().hex,
        )
        limit_2 = unit.new_limit_ref(
            project_id=self.project_bar['id'],
            service_id=self.service_one['id'],
            region_id=self.region_two['id'],
            resource_name='snapshot',
            resource_limit=5,
            id=uuid.uuid4().hex,
        )
        PROVIDERS.unified_limit_api.create_limits([limit_1, limit_2])

        expect_limit = 8
        limit_update = {'id': limit_1['id'], 'resource_limit': expect_limit}
        res = PROVIDERS.unified_limit_api.update_limit(
            limit_1['id'], limit_update
        )
        self.assertEqual(expect_limit, res['resource_limit'])

        # 'id' can be omitted in the update body
        limit_update = {'resource_limit': expect_limit}
        res = PROVIDERS.unified_limit_api.update_limit(
            limit_2['id'], limit_update
        )
        self.assertEqual(expect_limit, res['resource_limit'])

    def test_list_limits(self):
        # create two limits
        limit_1 = unit.new_limit_ref(
            project_id=self.project_bar['id'],
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            resource_limit=10,
            id=uuid.uuid4().hex,
            domain_id=None,
        )
        limit_2 = unit.new_limit_ref(
            project_id=self.project_bar['id'],
            service_id=self.service_one['id'],
            region_id=self.region_two['id'],
            resource_name='snapshot',
            resource_limit=5,
            id=uuid.uuid4().hex,
            domain_id=None,
        )
        PROVIDERS.unified_limit_api.create_limits([limit_1, limit_2])

        # list
        hints = driver_hints.Hints()
        hints.add_filter('project_id', self.project_bar['id'])
        limits = PROVIDERS.unified_limit_api.list_limits(hints)
        self.assertEqual(2, len(limits))
        for re in limits:
            if re['id'] == limit_1['id']:
                self.assertDictEqual(limit_1, re)
            if re['id'] == limit_2['id']:
                self.assertDictEqual(limit_2, re)

    def test_list_limit_by_limit(self):
        self.config_fixture.config(list_limit=1)
        # create two limits
        limit_1 = unit.new_limit_ref(
            project_id=self.project_bar['id'],
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            resource_limit=10,
            id=uuid.uuid4().hex,
            domain_id=None,
        )
        limit_2 = unit.new_limit_ref(
            project_id=self.project_bar['id'],
            service_id=self.service_one['id'],
            region_id=self.region_two['id'],
            resource_name='snapshot',
            resource_limit=5,
            id=uuid.uuid4().hex,
            domain_id=None,
        )
        PROVIDERS.unified_limit_api.create_limits([limit_1, limit_2])

        # list, limit is 1
        hints = driver_hints.Hints()
        limits = PROVIDERS.unified_limit_api.list_limits(hints=hints)
        self.assertEqual(1, len(limits))
        if limits[0]['id'] == limit_1['id']:
            self.assertDictEqual(limit_1, limits[0])
        else:
            self.assertDictEqual(limit_2, limits[0])

    def test_list_limit_by_filter(self):
        limit_1 = unit.new_limit_ref(
            project_id=self.project_bar['id'],
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            resource_limit=10,
            id=uuid.uuid4().hex,
            domain_id=None,
        )
        limit_2 = unit.new_limit_ref(
            project_id=self.project_baz['id'],
            service_id=self.service_one['id'],
            region_id=self.region_two['id'],
            resource_name='snapshot',
            resource_limit=10,
            id=uuid.uuid4().hex,
            domain_id=None,
        )
        limit_3 = unit.new_limit_ref(
            project_id=None,
            service_id=self.service_one['id'],
            region_id=self.region_two['id'],
            resource_name='snapshot',
            resource_limit=10,
            id=uuid.uuid4().hex,
            domain_id=self.domain_default['id'],
        )
        PROVIDERS.unified_limit_api.create_limits([limit_1, limit_2, limit_3])

        hints = driver_hints.Hints()
        hints.add_filter('service_id', self.service_one['id'])
        res = PROVIDERS.unified_limit_api.list_limits(hints)
        self.assertEqual(3, len(res))

        hints = driver_hints.Hints()
        hints.add_filter('region_id', self.region_one['id'])
        res = PROVIDERS.unified_limit_api.list_limits(hints)
        self.assertEqual(1, len(res))
        self.assertDictEqual(limit_1, res[0])

        hints = driver_hints.Hints()
        hints.add_filter('resource_name', 'backup')
        res = PROVIDERS.unified_limit_api.list_limits(hints)
        self.assertEqual(0, len(res))

        hints = driver_hints.Hints()
        hints.add_filter('project_id', self.project_bar['id'])
        res = PROVIDERS.unified_limit_api.list_limits(hints)
        self.assertEqual(1, len(res))

        hints = driver_hints.Hints()
        hints.add_filter('domain_id', self.domain_default['id'])
        res = PROVIDERS.unified_limit_api.list_limits(hints)
        self.assertEqual(1, len(res))

    def test_list_limit_by_multi_filter_with_project_id(self):
        limit_1 = unit.new_limit_ref(
            project_id=self.project_bar['id'],
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            resource_limit=10,
            id=uuid.uuid4().hex,
        )
        limit_2 = unit.new_limit_ref(
            project_id=self.project_baz['id'],
            service_id=self.service_one['id'],
            region_id=self.region_two['id'],
            resource_name='snapshot',
            resource_limit=10,
            id=uuid.uuid4().hex,
        )
        PROVIDERS.unified_limit_api.create_limits([limit_1, limit_2])

        hints = driver_hints.Hints()
        hints.add_filter('service_id', self.service_one['id'])
        hints.add_filter('project_id', self.project_bar['id'])
        res = PROVIDERS.unified_limit_api.list_limits(hints)
        self.assertEqual(1, len(res))

    def test_get_limit(self):
        # create two limits
        limit_1 = unit.new_limit_ref(
            project_id=self.project_bar['id'],
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            resource_limit=10,
            id=uuid.uuid4().hex,
            domain_id=None,
        )
        limit_2 = unit.new_limit_ref(
            project_id=self.project_bar['id'],
            service_id=self.service_one['id'],
            region_id=self.region_two['id'],
            resource_name='snapshot',
            resource_limit=5,
            id=uuid.uuid4().hex,
            domain_id=None,
        )
        PROVIDERS.unified_limit_api.create_limits([limit_1, limit_2])

        # show one
        res = PROVIDERS.unified_limit_api.get_limit(limit_2['id'])
        self.assertDictEqual(limit_2, res)

    def test_get_limit_returns_not_found(self):
        self.assertRaises(
            exception.LimitNotFound,
            PROVIDERS.unified_limit_api.get_limit,
            uuid.uuid4().hex,
        )

    def test_delete_limit(self):
        # create two limits
        limit_1 = unit.new_limit_ref(
            project_id=self.project_bar['id'],
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            resource_limit=10,
            id=uuid.uuid4().hex,
        )
        limit_2 = unit.new_limit_ref(
            project_id=self.project_bar['id'],
            service_id=self.service_one['id'],
            region_id=self.region_two['id'],
            resource_name='snapshot',
            resource_limit=5,
            id=uuid.uuid4().hex,
        )
        PROVIDERS.unified_limit_api.create_limits([limit_1, limit_2])
        # delete one
        PROVIDERS.unified_limit_api.delete_limit(limit_1['id'])
        # delete again
        self.assertRaises(
            exception.LimitNotFound,
            PROVIDERS.unified_limit_api.get_limit,
            limit_1['id'],
        )

    def test_delete_limit_returns_not_found(self):
        self.assertRaises(
            exception.LimitNotFound,
            PROVIDERS.unified_limit_api.delete_limit,
            uuid.uuid4().hex,
        )

    def test_delete_limit_project(self):
        # create two limits
        limit_1 = unit.new_limit_ref(
            project_id=self.project_bar['id'],
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            resource_limit=10,
            id=uuid.uuid4().hex,
        )
        limit_2 = unit.new_limit_ref(
            project_id=self.project_bar['id'],
            service_id=self.service_one['id'],
            region_id=self.region_two['id'],
            resource_name='snapshot',
            resource_limit=5,
            id=uuid.uuid4().hex,
        )
        PROVIDERS.unified_limit_api.create_limits([limit_1, limit_2])

        # delete a unrelated project, the limits should still be there.
        PROVIDERS.resource_api.delete_project(self.project_baz['id'])
        ref = PROVIDERS.unified_limit_api.list_limits()
        self.assertEqual(2, len(ref))

        # delete the referenced project, the limits should be deleted as well.
        PROVIDERS.resource_api.delete_project(self.project_bar['id'])
        ref = PROVIDERS.unified_limit_api.list_limits()
        self.assertEqual([], ref)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import fixture as config_fixture

from keystone.identity.backends import ldap
from keystone.tests.unit import core
from keystone.tests.unit.identity.backends import test_base
from keystone.tests.unit.ksfixtures import ldapdb


class TestIdentityDriver(core.BaseTestCase, test_base.IdentityDriverTests):
    allows_name_update = False
    allows_self_service_change_password = False
    expected_is_domain_aware = False
    expected_default_assignment_driver = 'sql'
    expected_is_sql = False
    expected_generates_uuids = False

    def setUp(self):
        super().setUp()

        config_fixture_ = self.useFixture(config_fixture.Config())
        config_fixture_.config(
            group='ldap',
            url='fake://memory',
            user='cn=Admin',
            password='password',
            suffix='cn=example,cn=com',
        )

        self.useFixture(ldapdb.LDAPDatabase())

        self.driver = ldap.Identity()

    def test_delete_user(self):
        self.skip_test_overrides('N/A: LDAP has no write support')

    def test_delete_user_no_user_exc(self):
        self.skip_test_overrides('N/A: LDAP has no write support')

    def test_delete_group(self):
        self.skip_test_overrides('N/A: LDAP has no write support')

    def test_delete_group_doesnt_exist_exc(self):
        self.skip_test_overrides('N/A: LDAP has no write support')

    def test_remove_user_from_group(self):
        self.skip_test_overrides('N/A: LDAP has no write support')

    def test_remove_user_from_group_not_in_group(self):
        self.skip_test_overrides('N/A: LDAP has no write support')

    def test_remove_user_from_group_no_user(self):
        self.skip_test_overrides('N/A: LDAP has no write support')

    def test_remove_user_from_group_no_group(self):
        self.skip_test_overrides('N/A: LDAP has no write support')



# Copyright 2018 Catalyst Cloud Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from keystone.conf import utils

expiration = cfg.IntOpt(
    'expiration',
    default=300,
    min=0,
    max=86400,
    help=utils.fmt(
        """
The amount of time that a receipt should remain valid (in seconds). This value
should always be very short, as it represents how long a user has to reattempt
auth with the missing auth methods.
"""
    ),
)

provider = cfg.StrOpt(
    'provider',
    default='fernet',
    help=utils.fmt(
        """
Entry point for the receipt provider in the `keystone.receipt.provider`
namespace. The receipt provider controls the receipt construction and
validation operations. Keystone includes just the `fernet` receipt provider for
now. `fernet` receipts do not need to be persisted at all, but require that you
run `keystone-manage fernet_setup` (also see the `keystone-manage
fernet_rotate` command).
"""
    ),
)

caching = cfg.BoolOpt(
    'caching',
    default=True,
    help=utils.fmt(
        """
Toggle for caching receipt creation and validation data. This has no effect
unless global caching is enabled, or if cache_on_issue is disabled as we only
cache receipts on issue.
"""
    ),
)

cache_time = cfg.IntOpt(
    'cache_time',
    default=300,
    min=0,
    help=utils.fmt(
        """
The number of seconds to cache receipt creation and validation data. This has
no effect unless both global and `[receipt] caching` are enabled.
"""
    ),
)

cache_on_issue = cfg.BoolOpt(
    'cache_on_issue',
    default=True,
    help=utils.fmt(
        """
Enable storing issued receipt data to receipt validation cache so that first
receipt validation doesn't actually cause full validation cycle. This option
has no effect unless global caching and receipt caching are enabled.
"""
    ),
)


GROUP_NAME = __name__.split('.')[-1]
ALL_OPTS = [expiration, provider, caching, cache_time, cache_on_issue]


def register_opts(conf):
    conf.register_opts(ALL_OPTS, group=GROUP_NAME)


def list_opts():
    return {GROUP_NAME: ALL_OPTS}
